//============================================================================
/* Copyright (c) The Technical University of Denmark
 * Author: Ojaswa Sharma
 * E-mail: os@imm.dtu.dk
 */
//============================================================================
#ifndef _SIGNED_DISTANCE_KERNELS_
#define _SIGNED_DISTANCE_KERNELS_

#include <cutil_math.h>
typedef unsigned int uint;

#define MIN(a, b) (a<b)?a:b
#define MAX(a, b) (a>b)?a:b
#define EPS 0.01


//max dist  - maximum possible distance value. Could be the longest diagonal of the full volume
__global__ void d_init_interface_BBox(cudaPitchedPtr d_volPPtr, cudaExtent logicalGridSize, uint xmin, uint xmax, uint ymin, uint ymax, uint zmin, uint zmax, cudaExtent subvolIndex, float max_dist, int subvolDim, int blockDim)
{
  unsigned int _w, _h, _d;
  unsigned int pitchz = logicalGridSize.width*logicalGridSize.height;

  _d = blockIdx.x/pitchz;
  _h = (blockIdx.x - (_d*pitchz))/logicalGridSize.width;
  _w = blockIdx.x - (_d*pitchz) - (_h*logicalGridSize.width);

  //compute coordinates local (wthin subvolume) and global (for the large volume) coordinates
  unsigned int l_x = blockDim*_w + threadIdx.x;
  unsigned int l_y = blockDim*_h + threadIdx.y;
  unsigned int l_z = blockDim*_d + threadIdx.z;

  if(l_x < 1 || l_x > (subvolDim-2) || l_y < 1 || l_y > (subvolDim-2) || l_z < 1 || l_z > (subvolDim-2)) return;

  // Reuse vars _w, _h, _d to store global coordinates of the voxel
  _w = l_x + __umul24(subvolDim-2, subvolIndex.width);
  _h = l_y + __umul24(subvolDim-2, subvolIndex.height);
  _d = l_z + __umul24(subvolDim-2, subvolIndex.depth);

  ///Use cuboid interface
  //find if this voxel lies outside, on or inside the interface...using 6 planes defining a cuboid
  //planes: [xmin, xmax, ymin, ymax, zmin, zmax]
  volatile float voxel_sign;
  if (_w > xmin && _w < xmax && _h > ymin && _h < ymax && _d > zmin && _d < zmax)
      voxel_sign = max_dist;
    else
      voxel_sign = -max_dist;
  voxel_sign = voxel_sign>0.0 ? -max_dist : max_dist;

  float *row = (float*)((char*)d_volPPtr.ptr + (l_z*d_volPPtr.ysize + l_y)*d_volPPtr.pitch);
  row[l_x] = voxel_sign;  
}

__global__ void d_init_interface_superEllipsoid(cudaPitchedPtr d_volPPtr, cudaExtent logicalGridSize, uint xmin, uint xmax, uint ymin, uint ymax, uint zmin, uint zmax, cudaExtent subvolIndex, float max_dist, int subvolDim, int blockDim, float superEllipsoidPower)
{
  unsigned int _w, _h, _d;
  unsigned int pitchz = logicalGridSize.width*logicalGridSize.height;

  _d = blockIdx.x/pitchz;
  _h = (blockIdx.x - (_d*pitchz))/logicalGridSize.width;
  _w = blockIdx.x - (_d*pitchz) - (_h*logicalGridSize.width);

  //compute coordinates local (wthin subvolume) and global (for the large volume) coordinates
  unsigned int l_x = blockDim*_w + threadIdx.x;
  unsigned int l_y = blockDim*_h + threadIdx.y;
  unsigned int l_z = blockDim*_d + threadIdx.z;

  if(l_x < 1 || l_x > (subvolDim-2) || l_y < 1 || l_y > (subvolDim-2) || l_z < 1 || l_z > (subvolDim-2)) return;

  // Reuse vars _w, _h, _d to store global coordinates of the voxel
  _w = l_x + __umul24(subvolDim-2, subvolIndex.width);
  _h = l_y + __umul24(subvolDim-2, subvolIndex.height);
  _d = l_z + __umul24(subvolDim-2, subvolIndex.depth);

  volatile float voxel_sign;
  ///Use Super-ellipsoidal interface (x-xc)^n/rc^n + (y-yc)^n/ry^n + (z-zc)^n/rz^n = 1; n even = 4 (here)
  volatile float3 center = make_float3((float)(xmax+xmin)/2.0, (float)(ymax+ymin)/2.0, (float)(zmax+zmin)/2.0);
  volatile float3 radius = make_float3((float)(xmax-xmin)/2.0, (float)(ymax-ymin)/2.0, (float)(zmax-zmin)/2.0);

  voxel_sign = powf((((float)_w - center.x)/radius.x), superEllipsoidPower) + powf((((float)_h - center.y)/radius.y), superEllipsoidPower) + powf((((float)_d - center.z)/radius.z), superEllipsoidPower) - 1.0;
  voxel_sign = voxel_sign>0.0 ? -max_dist : max_dist;

  float *row = (float*)((char*)d_volPPtr.ptr + (l_z*d_volPPtr.ysize + l_y)*d_volPPtr.pitch);
  row[l_x] = voxel_sign;  
}
#endif
